from __future__ import absolute_import
import drink
from markdown import markdown

def mark_it_down(txt):
    return markdown(txt, ['tables', 'fenced_code', 'codehilite(force_linenos=True)'])

class MarkdownEditor(drink.types._Editable):
    def html(self, caption, group):
        return drink.types._Editable.html(self, caption, group, '''
<script type="text/javascript" >
   $(document).ready(function() {
      $("#%(id)s").markItUp(mySettings);
   });
</script>
<textarea id="%(id)s" name="%(name)s" cols="80" rows="25">%(value)s</textarea>
    ''')

class MarkdownPage(drink.Page):
    content = "# You'r document title\n\n[Edit me](edit)"

    mime = "markdown"

    doc = "A markdown rendered page"

    js = ['/static/markitup/jquery.markitup.js', '/static/markitup/sets/markdown/set.js']

    css = ['/static/markitup/skins/markitup/style.css', '/static/markitup/sets/markdown/style.css']

    editable_fields = drink.Page.editable_fields.copy()
    editable_fields.update({
        'content': MarkdownEditor("Content"),
        'mime': drink.types.Text(),
    })

    def process(self):
        return mark_it_down(drink.request.forms.get('data'))

    def view(self):

        html = mark_it_down(self.content)

        return drink.template('main.html', obj=self,
             html=html, authenticated=drink.request.identity,
             # do not include js code, or css code, it's only for editing
             classes=self.classes,
             )

exported = {'Web page (markdown)': MarkdownPage}



from __future__ import absolute_import
from drink.config import config
from bottle import request, abort

__all__ = ['classes', 'get_object', 'init']

classes = {}

objects_to_load = [k for k, v in config.items('objects')]

def get_object(current, objpath):

    path_list = [p for p in objpath.split('/') if p]
    last_idx = len(path_list) - 1
    for i, elt in enumerate(path_list):

        if i == last_idx:
            # getting
            try:
                current = current[elt]
                if 'r' not in request.identity.access(current):
                    abort(401, 'Not authorized')
                    return
            except (KeyError, AttributeError, TypeError):
                try:
                    current = getattr(current, elt)
                except AttributeError:
                    raise AttributeError(elt)
            break # found a matching object
        else:
            # traversal
            try:
                current = current[elt]
                if 't' not in request.identity.access(current):
                    abort(401, 'Not authorized')
                    return
            except (KeyError, AttributeError):
                raise AttributeError(elt)
    return current

def init():
    for obj in objects_to_load:
        try:
            exec('from .%s import exported'%obj)
            classes.update(exported)
        except Exception:
            print "Unable to load %s, remove it from config file in [objects] section."%obj
            raise



from __future__ import absolute_import
import drink

class ObjectBrowser(drink.Page):
    mime = "search"

    doc = "A very basic object finder"

    classes = {}

    def __init__(self, name, rootpath):
        drink.Page.__init__(self, name, rootpath)
        self.lastlog = {}

    def query(self):
        yield '<html><body>'
        pat = drink.request.forms.get('pattern')
        items = []
        def _match(i, ignore_case=False):
            for attr in ('title', 'content', 'description'):
                v = getattr(i, attr, None)
                if isinstance(v, basestring):
                    if pat in v:
                        return True
            return False

        to_scan = drink.db.values()
        auth = drink.request.identity

        for item in to_scan:
            if 'r' in auth.access(item) and _match(item):
                items.append(item)
                yield '<li><a href="%s">%s</a></li>'%(item.path, item.title)
            try:
                to_scan.extend(item.itervalues())
            except AttributeError:
                pass
        self.lastlog[auth.id] = (pat, items)
        yield '<div>Done!</div><a href="%s">make new search...</a></body></html>'%self.path

    def view(self):

        auth = drink.request.identity

        if auth.id in self.lastlog:
            pat, items = self.lastlog[auth.id]
        else:
            pat = ''
            items = None


        form = ['<form class="query_form" id="query_form" action="query" method="post">',
            drink.types.Text('Look for').html('pattern', pat),
            '<input class="submit" type="submit" value="GO!"/></form>']
        if items:
            form.extend('<li><a href="%s">%s</a></li>'%(i.path, i.title) for i in items)

        return drink.template('main.html', obj=self, html='\n'.join(form), authenticated=auth, classes=self.classes)

exported = {"Finder": ObjectBrowser}



from __future__ import absolute_import
import os
from mimetypes import guess_type
from bottle import request, abort
from ZODB.blob import Blob
#from persistent.dict import PersistentDict
import persistent
from persistent.mapping import default
import transaction
from drink import template
import drink
from . import classes

# TODO: make Model inherit Persistent + Page inherit PersistentDict
#  Model & Page will share most of the interface but Model can't have any children (it's an end-point property)
# ??? Models will be self-editable, wilth _Editable-like interface, they may have a link...

class _Mock(object): pass

def get_type(filename):
    return guess_type(filename)[0] or 'application/octet-stream'

class Model(persistent.Persistent):

    # Dict-like methods

    def __init__(self, name, rootpath=None):
        persistent.Persistent.__init__(self)
        self.data = {}
        self.read_groups = set()
        self.write_groups = set()

        if not isinstance(name, basestring):
            # Root object special case, will pass the dict or None...
            if name is not None:
                self.update(name)
            name = '/'
            rootpath = '/'
            self.id = '.'
        else:
            if not name or name[0] in r'/.$%_':
                drink.abort(401, 'Wrong identifier: %r'%name)
            # minor sanity check
            self.id = name.replace(' ', '-').replace('\t','_').replace('/','.').replace('?', '')

        self.rootpath = rootpath

        if not hasattr(self, 'title'):
            self.title = name.replace('_', ' ').replace('-', ' ').capitalize()

        try:
            self.owner
        except AttributeError:
            self.owner = request.identity.user

    def __repr__(self):
        return repr(self.data)

    def __cmp__(self, dict):
        if isinstance(dict, Model):
            return cmp(self.data, dict.data)
        else:
            return cmp(self.data, dict)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, key):
        if key in self.data:
            return self.data[key]
        if hasattr(self.__class__, "__missing__"):
            return self.__class__.__missing__(self, key)
        raise KeyError(key)

    def __setitem__(self, key, item):
        if key in self:
            abort(401, "%r is already defined!"%key)
        self.data[key] = item
        self._p_changed = 1

    def __delitem__(self, key):
        del self.data[key]
        self._p_changed = 1

    def clear(self):
        self.data.clear()
        self._p_changed = 1

    def copy(self):
        if self.__class__ is Model:
            return Model(self.data.copy())
        import copy
        data = self.data
        try:
            self.data = {}
            c = copy.copy(self)
        finally:
            self.data = data
        c.update(self)
        return c

    def keys(self):
        return self.data.keys()

    def items(self):
        return self.data.items()

    def iteritems(self):
        return self.data.iteritems()

    def iterkeys(self):
        return self.data.iterkeys()

    def itervalues(self):
        return self.data.itervalues()

    def values(self):
        return self.data.values()

    def has_key(self, key):
        return key in self.data

    def update(self, dict=None, **kwargs):
        if dict is None:
            pass
        elif isinstance(dict, Model):
            self.data.update(dict.data)
        elif isinstance(dict, type({})) or not hasattr(dict, 'items'):
            self.data.update(dict)
        else:
            for k, v in dict.items():
                self[k] = v
        if len(kwargs):
            self.data.update(kwargs)
        self._p_changed = 1

    def get(self, key, failobj=None):
        if key not in self:
            return failobj
        return self[key]

    def setdefault(self, key, failobj=None):
        if key not in self:
            self[key] = failobj
        return self[key]

    def pop(self, key, *args):
        r = self.data.pop(key, *args)
        self._p_changed = 1
        return r

    def popitem(self):
        r = self.data.popitem()
        self._p_changed = 1
        return r

    def __contains__(self, key):
        return key in self.data

    @classmethod
    def fromkeys(cls, iterable, value=None):
        d = cls()
        for key in iterable:
            d[key] = value
        return d

    def __iter__(self):
        return iter(self.data)

    # Model properties

    editable_fields = {
        'title': drink.types.Text('Title')
    }

    owner_fields = {
        'read_groups':
            drink.types.GroupCheckBoxes("Read-enabled groups", group="x_permissions"),
        'min_rights':
            drink.types.Text("Every user's permissions (wrta)", group="x_permissions"),
        'write_groups':
            drink.types.GroupCheckBoxes("Write-enabled groups", group="x_permissions")
    }

    admin_fields = {
        #'id': drink.types.Id(),
    }

    css = None

    js = None

    html = None

    mime = 'model'

    classes = drink.classes

    min_rights = ''

    # Model methods

    def __hash__(self):
        return hash(self.id)

    def view(self):
        return "Not viewable"

    def struct(self, childs=True):

        a = request.identity.access

        d = dict()

        if 'r' in a(self):
            for k in self.editable_fields.keys():
                v = getattr(self, k, None)
                if isinstance(v, Model):
                    if not 'r' in a(v):
                        continue
                    v = v.struct(False)
                    v['id'] = k
                elif isinstance(v, (basestring, int, float)):
                    pass # serializes well in json
                else:
                    v = "N/A"
                d[k] = v


            if childs:
                it = [v.struct() for v in self.itervalues() if 'r' in a(v)]
                d['items'] = it

        d['id'] = self.id
        d['title'] = self.title
        d['path'] = self.rootpath
        d['mime'] = self.mime

        return d

    @property
    def path(self):
        return self.rootpath + self.id + '/'

    def edit(self, resume=None):
        r = resume or self._edit()
        transaction.commit() # commit before eventual redirect
        if callable(r[0]):
            return r[0](*r[1:])
        else:
            return r

    def _edit(self):
        if 'w' not in request.identity.access(self):
            return abort(401, "Not authorized")

        items = self.editable_fields.items()
        if request.identity.id == self.owner.id or request.identity.admin:
            items += self.owner_fields.items()

        if request.identity.admin:
            items += self.admin_fields.items()

        forms = request.forms

        if forms:
            editable = forms.keys()
            files = request.files.keys()

            for attr, caster in items:
                if attr in files:
                    caster.set(self, attr, request.files.get(attr))
                elif attr in editable:
                    caster.set(self, attr, forms.get(attr))

            transaction.commit()
            return (drink.rdr, self.path)
        else:
            if not items:
                form = ['<div class="error_message">Not editable, sorry...</div>']
            else:
                # sort by group+id
                items.sort(key=lambda o: o[1].group+o[0])
                current_group = None
                form_opts = []
                form = []
                for field, factory in items:
                    if factory.form_attr:
                        form_opts.append(factory.form_attr)
                    if factory.group != current_group:
                        old_group = current_group
                        current_group = factory.group
                        if old_group:
                            form.append('</div>')
                        form.append('<div class="%s_grp">'%current_group)
                    val = getattr(self, field, '')
                    form.append('<div class="input">%s</div>'%factory.html(field, val))
                form.append('</div><div class="buttons"><input class="submit" type="submit" value="Save changes please"/></div></form>')
                form.insert(0, '<form class="auto_edit_form" id="auto_edit_form" action="edit" %s method="post">'%(' '.join(form_opts)))
            return drink.template('main.html', obj=self, html='\n'.join(form), css=self.css, js=self.js, classes=self.classes, authenticated=request.identity)

class Page(Model):

    mime = 'page'

    doc = 'An abstract page'

    def upload(self):
        filename = request.GET.get('qqfile', 'uploaded')
        o = self._add(filename, 'File',
            request.identity.user.default_read_groups,
            request.identity.user.default_write_groups)
        fake_post_obj = _Mock()
        fake_post_obj.file = request.body
        fake_post_obj.filename = filename
        o.editable_fields['content'].set(o, 'content', fake_post_obj)
        o.mimetype = get_type(fake_post_obj.filename)
        return {'success': True, # for upload function
        # following describes the object:
        'path': o.rootpath, 'id': o.id, 'mime': 'page', 'title': o.title}

    def rm(self):
        name = request.GET.get('name')
        if not ('a' in request.identity.access(self) and 'w' in request.identity.access(self[name])):
            return abort(401, "Not authorized")
        try:
            parent_path = self.path
        except AttributeError: # XXX: unclean
            parent_path = '.'
        del self[name]
        transaction.commit()
        return drink.rdr(parent_path)

    def _add(self, name, cls, read_groups, write_groups):
        if isinstance(cls, basestring):
            klass = self.classes[cls] if self.classes else classes[cls]
        else:
            klass = cls
        new_obj = klass(name, self.path)
        if not new_obj.read_groups:
            new_obj.read_groups = set(read_groups)
        if not new_obj.write_groups:
            new_obj.write_groups = set(write_groups)

        self[new_obj.id] = new_obj

        transaction.commit()
        return new_obj

    def add(self, name=None, cls=None, read_groups=None, write_groups=None):
        auth = request.identity

        if 'a' not in auth.access(self):
            return abort(401, "Not authorized")
        name = name or request.GET.get('name')
        if None == cls:
            cls = request.GET.get('class')

        return drink.rdr(self._add(name, cls, auth.user.default_read_groups, auth.user.default_write_groups).rootpath)

    def view(self):
        return 'Not viewable...'


class ListPage(Page):
    doc = "An ordered folder-like display"

    mime = "folder"

    forced_order = None

    js = ['/static/listing.js']

    def __init__(self, name, rootpath=None):
        self.forced_order = []
        Page.__init__(self, name, rootpath)

    def iterkeys(self):
        k = Page.keys(self)
        return (x for x in self.forced_order if x in k) or k

    keys = iterkeys

    def move(self):
        self.forced_order = request.params.get('set').split('/')

    def itervalues(self):
        return (self[v] for v in self.keys())

    def reset_items(self):
        orig_order = []
        k = Page.keys(self)

        for item in self.forced_order:
            if item not in orig_order and item in k:
                orig_order.append(item)
        self.forced_order = orig_order
        return 'ok'

    def values(self):
        return list(self.itervalues)

    def __setitem__(self, name, val):
        if name is None:
            import pdb; pdb.set_trace()
        try:
            r = Page.__setitem__(self, name, val)
        except Exception:
            raise
        else:
            self.forced_order.append(name)
            return r

    def __delitem__(self, name):
        try:
            r = Page.__delitem__(self, name)
        except Exception:
            raise
        else:
            self.forced_order.remove(name)
            return r

    def view(self):
        return template('list.html', obj=self, css=self.css, js=self.js, classes=self.classes, authenticated=request.identity)

class StaticFile(Page):

    mime = "page"
    mimetype = "text/plain"
    doc = "A generic file"
    classes = {}

    owner_fields = {
        'read_groups':
            drink.types.GroupCheckBoxes("Read-enabled groups", group="x_permissions"),
        'write_groups':
            drink.types.GroupCheckBoxes("Write-enabled groups", group="x_permissions"),
        'mime': drink.types.Text(),
    }

    editable_fields = Page.editable_fields.copy()

    editable_fields.update( {
        'content':
            drink.types.File("File to upload"),
#        'content_name': drink.types.Text("File name"),
        'mimetype': drink.types.Text(),
    } )

    content = ''
    content_name = "unnamed"

    def raw(self):
        root, fname = os.path.split(self.content.committed())
        return drink.static_file(fname, root, mimetype=self.mimetype, download=self.content_name)
        #Alternatively: drink.response.headers['Content-Type'] = '...'
        # + read

    def _edit(self):
        # In case the user didn't change the mimetype
        # we guess it from the uploaded file name

        old_mimetype = self.mimetype
        r = Page._edit(self)
        if old_mimetype == self.mimetype \
           and request.forms.keys() and 'content' in request.files \
           and request.files['content'].filename:
            self.mimetype = get_type(self.content_name)
        return r

    def view(self):
        html = ['<div class="download"><a href="raw">Download file (original name: %r)</a></div>'%self.content_name]

        if self.content:
            mime = self.mimetype
            if mime.startswith('image/'):
                html.append('<img src="raw" />')
            elif mime in ('application/xml', ) or mime.startswith('text/'):
                f = self.content.open()
                html.append('<pre>')
                html.append(f.read())
                html.append('</pre>')

        return drink.template('main.html', obj=self, css=self.css, js=self.js, html='\n'.join(html), classes=self.classes, authenticated=request.identity)

exported = {'Folder index': ListPage, 'File': StaticFile}



mapping = {}
for obj in 'generic.Text generic.TextArea generic.ListPage tasks.TasksPage tasks.Task markdown.MarkdownPage'.split():
    namespace, klass = obj.rsplit('.', 1)
    mapping['flaskbox.objects.%s %s'%(namespace, klass)] = 'drink.objects.%s %s'%(namespace, klass)




__all__ = ['_Editable',
    'Text', 'TextArea', 'GroupListArea', 'GroupCheckBoxes',
    'Id', 'Int', 'Password', 'File']

import os
from bottle import request, abort
from persistent.dict import PersistentDict
from ZODB.blob import Blob
import transaction
from drink import template
import drink
from . import classes

class _Editable(object):

    form_attr = None

    def __init__(self, caption=None, group=None):
        self.caption = caption
        self.id = str(id(self))
        self.group = group if group else self.__class__.__name__

    def html(self, name, value, _template=None):
        d = self.__dict__.copy()
        d.update({'id': self.id, 'name': name, 'caption': self.caption or name, 'value': value})
        return ('<div class="editable"><label class="autoform" for="%(id)s">%(caption)s:</label>'+(_template or self._template)+'</div>')%d

    set = setattr


class Text(_Editable):

    _template = r'''<input type="text" size="%(size)d" id="%(id)s" value="%(value)s" name="%(name)s" />'''

    def __init__(self, caption=None, group=None, size=40):
        _Editable.__init__(self, caption, group)
        self.size = 40


class TextArea(_Editable):
    _template = '''<textarea rows="%(rows)s" cols="%(cols)s" id="%(id)s" name="%(name)s">%(value)s</textarea>'''

    def __init__(self, caption=None, group=None, rows=None, cols=None):
        _Editable.__init__(self, caption, group)
        self._rows = rows
        self._cols = cols

    def html(self, name, value):
        r = self._rows
        c = self._cols
        if isinstance(value, basestring):
            length = value.count('\n')
        else:
            length = len(value)

        if c is None: # automatic mode
            self.cols = 50
        else:
            self.cols = c

        if r is None:
            self.rows = min(400, max(10, length))
        else:
            self.rows = r

        return _Editable.html(self, name, value)

#class MultilineList(TextArea):
#    def html(self, name, value):
#        return TextArea.html(self, name, '\n'.join(value))
#
#    def set(self, obj, name, val):
#        setattr(obj, name, [line.strip() for line in val.split('\n') if line.strip()])

class GroupListArea(TextArea):
    def html(self, name, value):
        self._cols = 30
        self._rows = len(value) + 2
        return TextArea.html(self, name, '\n'.join(group.id for group in value))

    def set(self, obj, name, val):
        groups = drink.db.db['groups']
        setattr(obj, name, set(groups[line.strip()] for line in val.split('\n') if line.strip() in groups))


class GroupCheckBoxes(_Editable):

    def html(self, name, value):
        groups = [g for g in drink.db.db['groups']]
        values = [v.id for v in value]

        opts = [r'<input type="checkbox" name=%(name)s value="'+o+'" '+\
            ('checked="checked" />' if o in values else '/><span class="label">')+o+'</span>' for o in groups]
        return _Editable.html(self, name, value, '\n'.join(opts))


    def set(self, obj, name, val):
        groups = request.forms.getall(name)
        dgroups = drink.db.db['groups']
        setattr(obj, name, set(dgroups[g] for g in groups))


class Id(Text):
    def set(self, obj, name, val):
        parent = drink.get_object(drink.db, obj.rootpath)
        del parent[getattr(obj, name)]
        setattr(obj, name, val)
        parent[val] = obj

class Int(Text):
    def set(self, obj, name, val):
        setattr(obj, name, int(val))

class Password(Text):
    _template = r'''<input type="password" size="%(size)d" id="%(id)s" name="%(name)s" value="%(value)s" />'''

class File(_Editable):
    _template = '<input name="%(name)s" id="%(id)s" type="file" />'

    form_attr = 'enctype="multipart/form-data"'

    def set(self, obj, name, val):
        if val == '':
            return
        setattr(obj, name+"_name", val.filename)
        new_o = Blob()
        setattr(obj, name, new_o)
        o_fd = new_o.open('w')
        chunk_sz = 2**20 # 1MB
        while True:
            dat = val.file.read(chunk_sz)
            if not dat:
                break
            o_fd.write(dat)
        o_fd.close()



__all__ = ['BASE_DIR', 'config']

import os
import ConfigParser

BASE_DIR = os.path.abspath(os.path.split(__file__)[0])

config = ConfigParser.ConfigParser()

config.read([os.path.join(BASE_DIR, 'settings.ini'),
     os.path.expanduser('~/.drink.ini')])


import drink
from . import classes
import transaction
from hashlib import md5


class User(drink.Model):

    mime = "user"

    doc = "User object"

    password = ''

    email = ''

    classes = {}

    groups = set()

    default_read_groups = set()

    default_write_groups = set()

    admin_fields = drink.Model.admin_fields.copy()

    admin_fields.update( {
        'doc': drink.types.Text(),
        'groups': drink.types.GroupCheckBoxes(),
        'read_groups':
            drink.types.GroupCheckBoxes("Read-enabled groups", group="x_permissions"),
        'min_rights':
            drink.types.Text("Every user's permissions (wrta)", group="x_permissions"),
        'write_groups':
            drink.types.GroupCheckBoxes("Write-enabled groups", group="x_permissions")
    } )

    owner_fields = {
    # FIXME: Don't look ordered by group !
        'title': drink.types.Text('Nickname', group='0'),
        'name': drink.types.Text(group='1'),
        'surname': drink.types.Text(group='2'),
        'email': drink.types.Text(group='3'),
        'password': drink.types.Password(group='4'),
        'default_read_groups': drink.types.GroupCheckBoxes('Default readers Groups', group='x_permissions'),
        'default_write_groups': drink.types.GroupCheckBoxes('Default writers Groups', group='x_permissions'),
    }

    editable_fields = {
    }

    def __init__(self, name, rootpath):
        drink.Model.__init__(self, name, rootpath)
        name = self.id
        self.phones = {}
        self.groups = set()
        self.name = "no name"
        self.surname = "no surname"
        group_list = drink.db.db["groups"]
        group_list._add(name, Group, {}, {})
        mygroup = group_list[name]
        self.groups.add(mygroup)
        self.write_groups.add(mygroup)
        self.groups.add(group_list["users"])
        self.owner = self
        transaction.commit()

    def view(self):
        drink.rdr(self.path+'edit')

    def edit(self):
        r = drink.Model._edit(self)
        uid = md5(self.email).hexdigest()
        self.mime = 'http://www.gravatar.com/avatar/%s?s=32'%uid
        return drink.Model.edit(self, resume=r)


class UserList(drink.ListPage):
    doc = "Users folder"
    mime = "group"

    classes = {'User': User}


class Group(drink.Page):

    mime = "group"

    doc = "A group"

    name = "unnamed group"

    classes = {}

    editable_fields = {}

    def view(self):
        return 'keep out'


class GroupList(drink.ListPage):
    doc = "Groups"
    mime = "group"

    classes = {'Group': Group}


exported = {}



import setuptools
from setuptools import setup, find_packages

setup(
    name="drink",
    version="0.0.1",
    author="Fabien Devaux",
    author_email="fdev31@gmail.com",
    license="BSD",
    platform="all",
    description="my web sandbox (ZODB + Bottle)",
    long_description="""
    Drink
    =====

    Alpha program, base for a lightweight CMS / Intranet / etc...
    """,
    include_package_data=True,
    zip_safe=False,
    entry_points = {
        "console_scripts": [
            'drink = drink:startup',
            ],
        "setuptools.installation" : [
            'eggsecutable = drink:startup'
            ]
        },
    packages=find_packages(),
    install_requires = ['jinja2', 'markdown', 'ZODB3'],
)




from __future__ import absolute_import
import transaction
import time
from hashlib import sha1
import drink

class TODO(drink.Page):

    doc = "Something to do"

    editable_fields = {
        'title': drink.types.Text("Title", group="a"),
        'content': drink.types.TextArea("Summary", group="b"),
        #'date': drink.types.Date,
    }

    content = ''

    def view(self):
        drink.rdr(self.path+'edit')


class TODOList(drink.ListPage):

    doc = "A TODO list"

    classes = {'TODO': TODO}

    mime = "tasks"

    editable_fields = {
        'title': drink.types.Text("Title", group="a"),
    }

    html = """
    <div><a href="../"><img class="icon" src="/static/back.png" />Exit!</a></div>

    <h1>Daily tasks</h1>
    <div class="shoppingList">
        <ul>
        </ul>
    </div>
    """

exported = {'TODO list': TODOList}



from UserDict import IterableUserDict
import ZODB.config
import transaction
import drink
import atexit
import threading

class Database(object):
    def __init__(self, wsgi_app, config_file):
        self.app = wsgi_app
        self._config = config_file
        self.locals = threading.local()
        self._db = None
        self.connection = None
        wsgi_app.add_hook('before_request', self.start_request)
        wsgi_app.add_hook('after_request', self.close_request)
        atexit.register(self._cleanup)


    def _cleanup(self):
        if self._db:
            self._db.close()
        if getattr(self.locals, 'c', None):
            self.locals.c.close()
        self.app.hooks.clear()

    def __del__(self):
        self._cleanup()

    def pack(self):
        return self._db.pack()

    @property
    def db(self):
        if self._db is None:
            self._db = ZODB.config.databaseFromURL(self._config)

        r = getattr(self.locals, 'root', None)

        if None == r:
            c = self.locals.c = self._db.open()
            r = self.locals.root = c.root()

        return r

    def start_request(self):
        d = self.db
        transaction.begin()
        return d

    __enter__ = start_request

    def close_request(self):
        transaction.commit()

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_value:
            transaction.abort()
        else:
            transaction.commit()
